/* Copyright (c) 1994 The Geometry Center; University of Minnesota
   1300 South Second Street;  Minneapolis, MN  55454, USA;

This file is free software; you can redistribute it and/or modify it only
under the the terms of the GNU GENERAL PUBLIC LICENSE which should be
included along with this file.  This software may be obtained via anonymous
ftp from geom.umn.edu; email: software@geom.umn.edu. */

/* Author: Daeron Meyer */

void	UICreate();
void	UIEnd();
Window	MonitorWindow();
void    UISetwin(Window);
void    UIMinsize(int, int);
void    UISetFrame();
void    UISetColormap();
